# -*- coding: utf-8 -*-
"""Rigid body related equations.
"""
from pysph.base.reduce_array import parallel_reduce_array
from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
import numpy as np
import numpy
from math import sqrt


def skew(vec):
    import sympy as S
    x, y, z = vec[0], vec[1], vec[2]
    return S.Matrix([[0, -z, y], [z, 0, -x], [-y, x, 0]])


def get_alpha_dot():
    """Use sympy to perform most of the math and use the resulting formulae
    to calculate:

            inv(I) (\tau - w x (I w))
    """
    import sympy as S
    ixx, iyy, izz, ixy, ixz, iyz = S.symbols("ixx, iyy, izz, ixy, ixz, iyz")
    tx, ty, tz = S.symbols("tx, ty, tz")
    wx, wy, wz = S.symbols('wx, wy, wz')
    tau = S.Matrix([tx, ty, tz])
    I = S.Matrix([[ixx, ixy, ixz], [ixy, iyy, iyz], [ixz, iyz, izz]])
    w = S.Matrix([wx, wy, wz])
    Iinv = I.inv()
    Iinv.simplify()
    # inv(I) (\tau - w x (Iw))
    res = Iinv*(tau - w.cross(I*w))
    res.simplify()
    # Now do some awesome sympy magic.
    syms, result = S.cse(res, symbols=S.numbered_symbols('tmp'))
    for lhs, rhs in syms:
        print("%s = %s" % (lhs, rhs))
    for i in range(3):
        print("omega_dot[%d] =" % i, result[0][i])


def get_torque():
    """Use sympy to perform some simple math.
        R x F
        C_m x F
        w x r
    """
    import sympy as S
    x, y, z, fx, fy, fz = S.symbols("x, y, z, fx, fy, fz")
    R = S.Matrix([x, y, z])
    F = S.Matrix([fx, fy, fz])
    print("Torque:", R.cross(F))
    cx, cy, cz = S.symbols('cx, cy, cz')
    d = S.Matrix([cx, cy, cz])
    print("c_m x f = ", d.cross(F))
    wx, wy, wz = S.symbols('wx, wy, wz')
    rx, ry, rz = S.symbols('rx, ry, rz')
    w = S.Matrix([wx, wy, wz])
    r = S.Matrix([rx, ry, rz])
    print("w x r = %s" % w.cross(r))


# This is defined to silence editor warnings for the use of declare.
def declare(*args): pass


class RigidBodyMoments(Equation):
    def reduce(self, dst, t, dt):
        # FIXME: this will be slow in opencl
        nbody = declare('int')
        i = declare('int')
        base_mi = declare('int')
        base = declare('int')
        nbody = dst.num_body[0]
        if dst.gpu:
            dst.gpu.pull('omega', 'x', 'y', 'z', 'fx', 'fy', 'fz')

        d_mi = declare('object')
        m = declare('object')
        x = declare('object')
        y = declare('object')
        z = declare('object')
        fx = declare('object')
        fy = declare('object')
        fz = declare('object')
        d_mi = dst.mi
        cond = declare('object')
        for i in range(nbody):
            cond = dst.body_id == i
            base = i*16
            m = dst.m[cond]
            x = dst.x[cond]
            y = dst.y[cond]
            z = dst.z[cond]
            # Find the total_mass, center of mass and second moments.
            d_mi[base + 0] = numpy.sum(m)
            d_mi[base + 1] = numpy.sum(m*x)
            d_mi[base + 2] = numpy.sum(m*y)
            d_mi[base + 3] = numpy.sum(m*z)
            # Only do the lower triangle of values moments of inertia.
            d_mi[base + 4] = numpy.sum(m*(y*y + z*z))
            d_mi[base + 5] = numpy.sum(m*(x*x + z*z))
            d_mi[base + 6] = numpy.sum(m*(x*x + y*y))

            d_mi[base + 7] = -numpy.sum(m*x*y)
            d_mi[base + 8] = -numpy.sum(m*x*z)
            d_mi[base + 9] = -numpy.sum(m*y*z)

            # the total force and torque
            fx = dst.fx[cond]
            fy = dst.fy[cond]
            fz = dst.fz[cond]
            d_mi[base + 10] = numpy.sum(fx)
            d_mi[base + 11] = numpy.sum(fy)
            d_mi[base + 12] = numpy.sum(fz)

            # Calculate the torque and reduce it.
            d_mi[base + 13] = numpy.sum(y*fz - z*fy)
            d_mi[base + 14] = numpy.sum(z*fx - x*fz)
            d_mi[base + 15] = numpy.sum(x*fy - y*fx)

        # Reduce the temporary mi values in parallel across processors.
        d_mi[:] = parallel_reduce_array(dst.mi)

        # Set the reduced values.
        for i in range(nbody):
            base_mi = i*16
            base = i*3
            m = d_mi[base_mi + 0]
            dst.total_mass[i] = m
            cx = d_mi[base_mi + 1]/m
            cy = d_mi[base_mi + 2]/m
            cz = d_mi[base_mi + 3]/m
            dst.cm[base + 0] = cx
            dst.cm[base + 1] = cy
            dst.cm[base + 2] = cz

            # The actual moment of inertia about center of mass from parallel
            # axes theorem.
            ixx = d_mi[base_mi + 4] - (cy*cy + cz*cz)*m
            iyy = d_mi[base_mi + 5] - (cx*cx + cz*cz)*m
            izz = d_mi[base_mi + 6] - (cx*cx + cy*cy)*m
            ixy = d_mi[base_mi + 7] + cx*cy*m
            ixz = d_mi[base_mi + 8] + cx*cz*m
            iyz = d_mi[base_mi + 9] + cy*cz*m

            d_mi[base_mi + 0] = ixx
            d_mi[base_mi + 1] = ixy
            d_mi[base_mi + 2] = ixz
            d_mi[base_mi + 3] = ixy
            d_mi[base_mi + 4] = iyy
            d_mi[base_mi + 5] = iyz
            d_mi[base_mi + 6] = ixz
            d_mi[base_mi + 7] = iyz
            d_mi[base_mi + 8] = izz

            fx = d_mi[base_mi + 10]
            fy = d_mi[base_mi + 11]
            fz = d_mi[base_mi + 12]
            dst.force[base + 0] = fx
            dst.force[base + 1] = fy
            dst.force[base + 2] = fz

            # Acceleration of CM.
            dst.ac[base + 0] = fx/m
            dst.ac[base + 1] = fy/m
            dst.ac[base + 2] = fz/m

            # Find torque about the Center of Mass and not origin.
            tx = d_mi[base_mi + 13]
            ty = d_mi[base_mi + 14]
            tz = d_mi[base_mi + 15]
            tx -= cy*fz - cz*fy
            ty -= -cx*fz + cz*fx
            tz -= cx*fy - cy*fx
            dst.torque[base + 0] = tx
            dst.torque[base + 1] = ty
            dst.torque[base + 2] = tz

            wx = dst.omega[base + 0]
            wy = dst.omega[base + 1]
            wz = dst.omega[base + 2]
            # Find omega_dot from: omega_dot = inv(I) (\tau - w x (Iw))
            # This was done using the sympy code above.
            tmp0 = iyz**2
            tmp1 = ixy**2
            tmp2 = ixz**2
            tmp3 = ixx*iyy
            tmp4 = ixy*ixz
            tmp5 = 1./(ixx*tmp0 + iyy*tmp2 - 2*iyz*tmp4 + izz*tmp1 - izz*tmp3)
            tmp6 = ixy*izz - ixz*iyz
            tmp7 = ixz*wx + iyz*wy + izz*wz
            tmp8 = ixx*wx + ixy*wy + ixz*wz
            tmp9 = tmp7*wx - tmp8*wz + ty
            tmp10 = ixy*iyz - ixz*iyy
            tmp11 = ixy*wx + iyy*wy + iyz*wz
            tmp12 = -tmp11*wx + tmp8*wy + tz
            tmp13 = tmp11*wz - tmp7*wy + tx
            tmp14 = ixx*iyz - tmp4
            dst.omega_dot[base + 0] = tmp5*(-tmp10*tmp12 -
                                            tmp13*(iyy*izz - tmp0) + tmp6*tmp9)
            dst.omega_dot[base + 1] = tmp5*(tmp12*tmp14 +
                                            tmp13*tmp6 - tmp9*(ixx*izz - tmp2))
            dst.omega_dot[base + 2] = tmp5*(-tmp10*tmp13 -
                                            tmp12*(-tmp1 + tmp3) + tmp14*tmp9)
        if dst.gpu:
            dst.gpu.push(
                'total_mass', 'mi', 'cm', 'force', 'ac', 'torque',
                'omega_dot'
            )


class RigidBodyMotion(Equation):
    def initialize(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w,
                   d_cm, d_vc, d_ac, d_omega, d_body_id):
        base = declare('int')
        base = d_body_id[d_idx]*3
        wx = d_omega[base + 0]
        wy = d_omega[base + 1]
        wz = d_omega[base + 2]
        rx = d_x[d_idx] - d_cm[base + 0]
        ry = d_y[d_idx] - d_cm[base + 1]
        rz = d_z[d_idx] - d_cm[base + 2]

        d_u[d_idx] = d_vc[base + 0] + wy*rz - wz*ry
        d_v[d_idx] = d_vc[base + 1] + wz*rx - wx*rz
        d_w[d_idx] = d_vc[base + 2] + wx*ry - wy*rx


class BodyForce(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(BodyForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_m, d_fx, d_fy, d_fz):
        d_fx[d_idx] = d_m[d_idx]*self.gx
        d_fy[d_idx] = d_m[d_idx]*self.gy
        d_fz[d_idx] = d_m[d_idx]*self.gz


class SummationDensityBoundary(Equation):
    r"""Equation to find the density of the
    fluid particle due to any boundary or a rigid body

    :math:`\rho_a = \sum_b {\rho}_fluid V_b W_{ab}`

    """
    def __init__(self, dest, sources, fluid_rho=1000.0):
        self.fluid_rho = fluid_rho
        super(SummationDensityBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, s_idx, s_m, s_V, WIJ):
        d_rho[d_idx] += self.fluid_rho * s_V[s_idx] * WIJ


class NumberDensity(Equation):
    def initialize(self, d_idx, d_V):
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_V, WIJ):
        d_V[d_idx] += WIJ


class SummationDensityRigidBody(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(SummationDensityRigidBody, self).__init__(dest, sources)

    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_rho, s_idx, s_V, WIJ):
        d_rho[d_idx] += self.rho0/s_V[s_idx]*WIJ


class ViscosityRigidBody(Equation):

    """The viscous acceleration on the fluid/solid due to a boundary.
    Implemented from Akinci et al. http://dx.doi.org/10.1145/2185520.2185558

    Use this with the fluid as a destination and body as source.
    """

    def __init__(self, dest, sources, rho0, nu):
        self.nu = nu
        self.rho0 = rho0
        super(ViscosityRigidBody, self).__init__(dest, sources)

    def loop(self, d_idx, d_m, d_au, d_av, d_aw, d_rho,
             s_idx, s_V, s_fx, s_fy, s_fz,
             EPS, VIJ, XIJ, R2IJ, DWIJ):
        phi_b = self.rho0/(s_V[s_idx]*d_rho[d_idx])
        vijdotxij = min(VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2], 0.0)

        fac = self.nu*phi_b*vijdotxij/(R2IJ + EPS)
        ax = fac*DWIJ[0]
        ay = fac*DWIJ[1]
        az = fac*DWIJ[2]
        d_au[d_idx] += ax
        d_av[d_idx] += ay
        d_aw[d_idx] += az
        s_fx[s_idx] += -d_m[d_idx]*ax
        s_fy[s_idx] += -d_m[d_idx]*ay
        s_fz[s_idx] += -d_m[d_idx]*az


class PressureRigidBody(Equation):

    """The pressure acceleration on the fluid/solid due to a boundary.
    Implemented from Akinci et al. http://dx.doi.org/10.1145/2185520.2185558

    Use this with the fluid as a destination and body as source.
    """

    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(PressureRigidBody, self).__init__(dest, sources)

    def loop(self, d_idx, d_m, d_rho, d_au, d_av, d_aw,  d_p,
             s_idx, s_V, s_fx, s_fy, s_fz, DWIJ):
        rho1 = 1.0/d_rho[d_idx]
        fac = -d_p[d_idx]*rho1*rho1*self.rho0/s_V[s_idx]
        ax = fac*DWIJ[0]
        ay = fac*DWIJ[1]
        az = fac*DWIJ[2]
        d_au[d_idx] += ax
        d_av[d_idx] += ay
        d_aw[d_idx] += az
        s_fx[s_idx] += -d_m[d_idx]*ax
        s_fy[s_idx] += -d_m[d_idx]*ay
        s_fz[s_idx] += -d_m[d_idx]*az


class AkinciRigidFluidCoupling(Equation):
    """Force between a solid sphere and a SPH fluid particle.  This is
    implemented using Akinci's[1] force and additional force from solid
    bodies pressure which is implemented by Liu[2]

    [1]'Versatile Rigid-Fluid Coupling for Incompressible SPH'

    URL: https://graphics.ethz.ch/~sobarbar/papers/Sol12/Sol12.pdf

    [2]A 3D Simulation of a Moving Solid in Viscous Free-Surface Flows by
    Coupling SPH and DEM

    https://doi.org/10.1155/2017/3174904


    Note: Here forces for both the phases are added at once.
          Please make sure that this force is applied only once
          for both the particle properties.

    """
    def __init__(self, dest, sources, fluid_rho=1000):
        super(AkinciRigidFluidCoupling, self).__init__(dest, sources)
        self.fluid_rho = fluid_rho

    def loop(self, d_idx, d_m, d_rho, d_au, d_av, d_aw,  d_p,
             s_idx, s_V, s_fx, s_fy, s_fz, DWIJ, s_m, s_p, s_rho):

        psi = s_V[s_idx] * self.fluid_rho

        _t1 = 2 * d_p[d_idx] / (d_rho[d_idx]**2)

        d_au[d_idx] += -psi * _t1 * DWIJ[0]
        d_av[d_idx] += -psi * _t1 * DWIJ[1]
        d_aw[d_idx] += -psi * _t1 * DWIJ[2]

        s_fx[s_idx] += d_m[d_idx] * psi * _t1 * DWIJ[0]
        s_fy[s_idx] += d_m[d_idx] * psi * _t1 * DWIJ[1]
        s_fz[s_idx] += d_m[d_idx] * psi * _t1 * DWIJ[2]


class LiuFluidForce(Equation):
    """Force between a solid sphere and a SPH fluid particle.  This is
    implemented using Akinci's[1] force and additional force from solid
    bodies pressure which is implemented by Liu[2]

    [1]'Versatile Rigid-Fluid Coupling for Incompressible SPH'

    URL: https://graphics.ethz.ch/~sobarbar/papers/Sol12/Sol12.pdf

    [2]A 3D Simulation of a Moving Solid in Viscous Free-Surface Flows by
    Coupling SPH and DEM

    https://doi.org/10.1155/2017/3174904


    Note: Here forces for both the phases are added at once.
          Please make sure that this force is applied only once
          for both the particle properties.

    """
    def __init__(self, dest, sources):
        super(LiuFluidForce, self).__init__(dest, sources)

    def loop(self, d_idx, d_m, d_rho, d_au, d_av, d_aw,  d_p,
             s_idx, s_V, s_fx, s_fy, s_fz, DWIJ, s_m, s_p, s_rho):
        _t1 = s_p[s_idx] / (s_rho[s_idx]**2) + d_p[d_idx] / (d_rho[d_idx]**2)

        d_au[d_idx] += -s_m[s_idx] * _t1 * DWIJ[0]
        d_av[d_idx] += -s_m[s_idx] * _t1 * DWIJ[1]
        d_aw[d_idx] += -s_m[s_idx] * _t1 * DWIJ[2]

        s_fx[s_idx] += d_m[d_idx] * s_m[s_idx] * _t1 * DWIJ[0]
        s_fy[s_idx] += d_m[d_idx] * s_m[s_idx] * _t1 * DWIJ[1]
        s_fz[s_idx] += d_m[d_idx] * s_m[s_idx] * _t1 * DWIJ[2]


class RigidBodyForceGPUGems(Equation):
    """This is inspired from
    http://http.developer.nvidia.com/GPUGems3/gpugems3_ch29.html
    and
    BK Mishra's article on DEM
    http://dx.doi.org/10.1016/S0301-7516(03)00032-2
    A review of computer simulation of tumbling mills by the discrete element
    method: Part I - contact mechanics
    """
    def __init__(self, dest, sources, k=1.0, d=1.0, eta=1.0, kt=1.0):
        """Note that d is a factor multiplied with the "h" of the particle.
        """
        self.k = k
        self.d = d
        self.eta = eta
        self.kt = kt
        super(RigidBodyForceGPUGems, self).__init__(dest, sources)

    def loop(self, d_idx, d_fx, d_fy, d_fz, d_h, d_total_mass, XIJ,
             RIJ, R2IJ, VIJ):
        vijdotrij = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]
        if RIJ > 1e-9:
            vijdotrij_r2ij = vijdotrij/R2IJ
            nij_x = XIJ[0]/RIJ
            nij_y = XIJ[1]/RIJ
            nij_z = XIJ[2]/RIJ
        else:
            vijdotrij_r2ij = 0.0
            nij_x = 0.0
            nij_y = 0.0
            nij_z = 0.0
        vijt_x = VIJ[0] - vijdotrij_r2ij*XIJ[0]
        vijt_y = VIJ[1] - vijdotrij_r2ij*XIJ[1]
        vijt_z = VIJ[2] - vijdotrij_r2ij*XIJ[2]

        d = self.d*d_h[d_idx]
        fac = self.k*d_total_mass[0]/d*max(d - RIJ, 0.0)

        d_fx[d_idx] += fac*nij_x - self.eta*VIJ[0] - self.kt*vijt_x
        d_fy[d_idx] += fac*nij_y - self.eta*VIJ[1] - self.kt*vijt_y
        d_fz[d_idx] += fac*nij_z - self.eta*VIJ[2] - self.kt*vijt_z


class RigidBodyCollision(Equation):
    """Force between two spheres is implemented using DEM contact force law.

    Refer https://doi.org/10.1016/j.powtec.2011.09.019 for more
    information.

    Open-source MFIX-DEM software for gas–solids flows:
    Part I—Verification studies .

    """
    def __init__(self, dest, sources, kn=1e3, mu=0.5, en=0.8):
        """Initialise the required coefficients for force calculation.


        Keyword arguments:
        kn -- Normal spring stiffness (default 1e3)
        mu -- friction coefficient (default 0.5)
        en -- coefficient of restitution (0.8)

        Given these coefficients, tangential spring stiffness, normal and
        tangential damping coefficient are calculated by default.

        """
        self.kn = kn
        self.kt = 2. / 7. * kn
        m_eff = np.pi * 0.5**2 * 1e-6 * 2120
        self.gamma_n = -(2 * np.sqrt(kn * m_eff) * np.log(en)) / (
            np.sqrt(np.pi**2 + np.log(en)**2))
        self.gamma_t = 0.5 * self.gamma_n
        self.mu = mu
        super(RigidBodyCollision, self).__init__(dest, sources)

    def loop(self, d_idx, d_fx, d_fy, d_fz, d_h, d_total_mass, d_rad_s,
             d_tang_disp_x, d_tang_disp_y, d_tang_disp_z, d_tang_velocity_x,
             d_tang_velocity_y, d_tang_velocity_z, s_idx, s_rad_s, XIJ, RIJ,
             R2IJ, VIJ):
        overlap = 0
        if RIJ > 1e-9:
            overlap = d_rad_s[d_idx] + s_rad_s[s_idx] - RIJ

        if overlap > 0:
            # normal vector passing from particle i to j
            nij_x = -XIJ[0] / RIJ
            nij_y = -XIJ[1] / RIJ
            nij_z = -XIJ[2] / RIJ

            # overlap speed: a scalar
            vijdotnij = VIJ[0] * nij_x + VIJ[1] * nij_y + VIJ[2] * nij_z

            # normal velocity
            vijn_x = vijdotnij * nij_x
            vijn_y = vijdotnij * nij_y
            vijn_z = vijdotnij * nij_z

            # normal force with conservative and dissipation part
            fn_x = -self.kn * overlap * nij_x - self.gamma_n * vijn_x
            fn_y = -self.kn * overlap * nij_y - self.gamma_n * vijn_y
            fn_z = -self.kn * overlap * nij_z - self.gamma_n * vijn_z

            # ----------------------Tangential force---------------------- #

            # tangential velocity
            d_tang_velocity_x[d_idx] = VIJ[0] - vijn_x
            d_tang_velocity_y[d_idx] = VIJ[1] - vijn_y
            d_tang_velocity_z[d_idx] = VIJ[2] - vijn_z

            dtvx = d_tang_velocity_x[d_idx]
            dtvy = d_tang_velocity_y[d_idx]
            dtvz = d_tang_velocity_z[d_idx]
            _tang = sqrt(dtvx*dtvx + dtvy*dtvy + dtvz*dtvz)

            # tangential unit vector
            tij_x = 0
            tij_y = 0
            tij_z = 0
            if _tang > 0:
                tij_x = d_tang_velocity_x[d_idx] / _tang
                tij_y = d_tang_velocity_y[d_idx] / _tang
                tij_z = d_tang_velocity_z[d_idx] / _tang

            # damping force or dissipation
            ft_x_d = -self.gamma_t * d_tang_velocity_x[d_idx]
            ft_y_d = -self.gamma_t * d_tang_velocity_y[d_idx]
            ft_z_d = -self.gamma_t * d_tang_velocity_z[d_idx]

            # tangential spring force
            ft_x_s = -self.kt * d_tang_disp_x[d_idx]
            ft_y_s = -self.kt * d_tang_disp_y[d_idx]
            ft_z_s = -self.kt * d_tang_disp_z[d_idx]

            ft_x = ft_x_d + ft_x_s
            ft_y = ft_y_d + ft_y_s
            ft_z = ft_z_d + ft_z_s

            # coulomb law
            ftij = sqrt((ft_x**2) + (ft_y**2) + (ft_z**2))
            fnij = sqrt((fn_x**2) + (fn_y**2) + (fn_z**2))

            _fnij = self.mu * fnij

            if _fnij < ftij:
                ft_x = -_fnij * tij_x
                ft_y = -_fnij * tij_y
                ft_z = -_fnij * tij_z

            d_fx[d_idx] += fn_x + ft_x
            d_fy[d_idx] += fn_y + ft_y
            d_fz[d_idx] += fn_z + ft_z
        else:
            d_tang_velocity_x[d_idx] = 0
            d_tang_velocity_y[d_idx] = 0
            d_tang_velocity_z[d_idx] = 0

            d_tang_disp_x[d_idx] = 0
            d_tang_disp_y[d_idx] = 0
            d_tang_disp_z[d_idx] = 0


class RigidBodyWallCollision(Equation):
    """Force between sphere and a wall is implemented using
    DEM contact force law.

    Refer https://doi.org/10.1016/j.powtec.2011.09.019 for more
    information.

    Open-source MFIX-DEM software for gas–solids flows:
    Part I—Verification studies .

    """
    def __init__(self, dest, sources, kn=1e3, mu=0.5, en=0.8):
        """Initialise the required coefficients for force calculation.


        Keyword arguments:
        kn -- Normal spring stiffness (default 1e3)
        mu -- friction coefficient (default 0.5)
        en -- coefficient of restitution (0.8)

        Given these coefficients, tangential spring stiffness, normal and
        tangential damping coefficient are calculated by default.

        """
        self.kn = kn
        self.kt = 2. / 7. * kn
        m_eff = np.pi * 0.5**2 * 1e-6 * 2120
        self.gamma_n = -(2 * np.sqrt(kn * m_eff) * np.log(en)) / (
            np.sqrt(np.pi**2 + np.log(en)**2))
        print(self.gamma_n)
        self.gamma_t = 0.5 * self.gamma_n
        self.mu = mu
        super(RigidBodyWallCollision, self).__init__(dest, sources)

    def loop(self, d_idx, d_fx, d_fy, d_fz, d_h, d_total_mass, d_rad_s,
             d_tang_disp_x, d_tang_disp_y, d_tang_disp_z, d_tang_velocity_x,
             d_tang_velocity_y, d_tang_velocity_z, s_idx, XIJ, RIJ,
             R2IJ, VIJ, s_nx, s_ny, s_nz):
        # check overlap amount
        overlap = d_rad_s[d_idx] - (XIJ[0] * s_nx[s_idx] + XIJ[1] *
                                    s_ny[s_idx] + XIJ[2] * s_nz[s_idx])

        if overlap > 0:
            # basic variables: normal vector
            nij_x = -s_nx[s_idx]
            nij_y = -s_ny[s_idx]
            nij_z = -s_nz[s_idx]

            # overlap speed: a scalar
            vijdotnij = VIJ[0] * nij_x + VIJ[1] * nij_y + VIJ[2] * nij_z

            # normal velocity
            vijn_x = vijdotnij * nij_x
            vijn_y = vijdotnij * nij_y
            vijn_z = vijdotnij * nij_z

            # normal force with conservative and dissipation part
            fn_x = -self.kn * overlap * nij_x - self.gamma_n * vijn_x
            fn_y = -self.kn * overlap * nij_y - self.gamma_n * vijn_y
            fn_z = -self.kn * overlap * nij_z - self.gamma_n * vijn_z

            # ----------------------Tangential force---------------------- #

            # tangential velocity
            d_tang_velocity_x[d_idx] = VIJ[0] - vijn_x
            d_tang_velocity_y[d_idx] = VIJ[1] - vijn_y
            d_tang_velocity_z[d_idx] = VIJ[2] - vijn_z

            _tang = (
                (d_tang_velocity_x[d_idx]**2) + (d_tang_velocity_y[d_idx]**2) +
                (d_tang_velocity_z[d_idx]**2))**(1. / 2.)

            # tangential unit vector
            tij_x = 0
            tij_y = 0
            tij_z = 0
            if _tang > 0:
                tij_x = d_tang_velocity_x[d_idx] / _tang
                tij_y = d_tang_velocity_y[d_idx] / _tang
                tij_z = d_tang_velocity_z[d_idx] / _tang

            # damping force or dissipation
            ft_x_d = -self.gamma_t * d_tang_velocity_x[d_idx]
            ft_y_d = -self.gamma_t * d_tang_velocity_y[d_idx]
            ft_z_d = -self.gamma_t * d_tang_velocity_z[d_idx]

            # tangential spring force
            ft_x_s = -self.kt * d_tang_disp_x[d_idx]
            ft_y_s = -self.kt * d_tang_disp_y[d_idx]
            ft_z_s = -self.kt * d_tang_disp_z[d_idx]

            ft_x = ft_x_d + ft_x_s
            ft_y = ft_y_d + ft_y_s
            ft_z = ft_z_d + ft_z_s

            # coulomb law
            ftij = ((ft_x**2) + (ft_y**2) + (ft_z**2))**(1. / 2.)
            fnij = ((fn_x**2) + (fn_y**2) + (fn_z**2))**(1. / 2.)

            _fnij = self.mu * fnij

            if _fnij < ftij:
                ft_x = -_fnij * tij_x
                ft_y = -_fnij * tij_y
                ft_z = -_fnij * tij_z

            d_fx[d_idx] += fn_x + ft_x
            d_fy[d_idx] += fn_y + ft_y
            d_fz[d_idx] += fn_z + ft_z
            # print(d_fz[d_idx])
        else:
            d_tang_velocity_x[d_idx] = 0
            d_tang_velocity_y[d_idx] = 0
            d_tang_velocity_z[d_idx] = 0

            d_tang_disp_x[d_idx] = 0
            d_tang_disp_y[d_idx] = 0
            d_tang_disp_z[d_idx] = 0


class EulerStepRigidBody(IntegratorStep):
    """Fast but inaccurate integrator. Use this for testing"""
    def initialize(self):
        pass

    def stage1(self, d_idx, d_u, d_v, d_w, d_x, d_y, d_z,
               d_omega, d_omega_dot, d_vc, d_ac, d_num_body,
               dt=0.0):
        _i = declare('int')
        _j = declare('int')
        base = declare('int')
        if d_idx == 0:
            for _i in range(d_num_body[0]):
                base = 3*_i
                for _j in range(3):
                    d_vc[base + _j] += d_ac[base + _j]*dt
                    d_omega[base + _j] += d_omega_dot[base + _j]*dt

        d_x[d_idx] += dt*d_u[d_idx]
        d_y[d_idx] += dt*d_v[d_idx]
        d_z[d_idx] += dt*d_w[d_idx]


class RK2StepRigidBody(IntegratorStep):
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                   d_omega, d_omega0, d_vc, d_vc0, d_num_body):
        _i = declare('int')
        _j = declare('int')
        base = declare('int')
        if d_idx == 0:
            for _i in range(d_num_body[0]):
                base = 3*_i
                for _j in range(3):
                    d_vc0[base + _j] = d_vc[base + _j]
                    d_omega0[base + _j] = d_omega[base + _j]

        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

    def stage1(self, d_idx, d_u, d_v, d_w, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_omega, d_omega_dot, d_vc, d_ac, d_omega0, d_vc0, d_num_body,
               dt=0.0):
        dtb2 = 0.5*dt
        _i = declare('int')
        j = declare('int')
        base = declare('int')
        if d_idx == 0:
            for _i in range(d_num_body[0]):
                base = 3*_i
                for j in range(3):
                    d_vc[base + j] = d_vc0[base + j] + d_ac[base + j]*dtb2
                    d_omega[base + j] = (d_omega0[base + j] +
                                         d_omega_dot[base + j]*dtb2)

        d_x[d_idx] = d_x0[d_idx] + dtb2*d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2*d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2*d_w[d_idx]

    def stage2(self, d_idx, d_u, d_v, d_w, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_omega, d_omega_dot, d_vc, d_ac, d_omega0, d_vc0, d_num_body,
               dt=0.0):
        _i = declare('int')
        j = declare('int')
        base = declare('int')
        if d_idx == 0:
            for _i in range(d_num_body[0]):
                base = 3*_i
                for j in range(3):
                    d_vc[base + j] = d_vc0[base + j] + d_ac[base + j]*dt
                    d_omega[base + j] = (d_omega0[base + j] +
                                         d_omega_dot[base + j]*dt)

        d_x[d_idx] = d_x0[d_idx] + dt*d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt*d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt*d_w[d_idx]



import numpy as np
from pysph.base.utils import get_particle_array_wcsph
x, y = np.mgrid[-1:1:50j, -1:1:50j]
mask = x*x + y*y < 1.0
pa = get_particle_array_wcsph(name='fluid', x=x[mask], y=y[mask])
plt.scatter(pa.x, pa.y, marker='.')



# This is file is generated from interpolate.py.mako
# Do not edit this file
from pysph.sph.equation import Equation
from compyle.api import declare
from pysph.sph.wc.linalg import gj_solve, augmented_matrix


class EvaluateUhat(Equation):
    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(EvaluateUhat, self).__init__(dest, sources)

    def initialize(self, d_idx, d_uho, d_Buh, d_vho, d_Bvh, d_who, d_Bwh):
        i = declare('int')

        for i in range(3):
            d_uho[4*d_idx+i] = 0.0
            d_Buh[4*d_idx+i] = 0.0
            d_vho[4*d_idx+i] = 0.0
            d_Bvh[4*d_idx+i] = 0.0
            d_who[4*d_idx+i] = 0.0
            d_Bwh[4*d_idx+i] = 0.0

    def loop(self, d_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y, d_z, s_rho,
             s_m, s_idx, XIJ, DWIJ,
             WIJ, s_uhat, d_Buh, s_vhat, d_Bvh, s_what, d_Bwh):
        Vj = s_m[s_idx] / s_rho[s_idx]
        uhj = s_uhat[s_idx]
        vhj = s_vhat[s_idx]
        whj = s_what[s_idx]
        i4 = declare('int')
        i4 = 4*d_idx

        d_Buh[i4+0] += uhj * WIJ * Vj
        d_Buh[i4+1] += uhj * DWIJ[0] * Vj
        d_Buh[i4+2] += uhj * DWIJ[1] * Vj
        d_Buh[i4+3] += uhj * DWIJ[2] * Vj
        d_Bvh[i4+0] += vhj * WIJ * Vj
        d_Bvh[i4+1] += vhj * DWIJ[0] * Vj
        d_Bvh[i4+2] += vhj * DWIJ[1] * Vj
        d_Bvh[i4+3] += vhj * DWIJ[2] * Vj
        d_Bwh[i4+0] += whj * WIJ * Vj
        d_Bwh[i4+1] += whj * DWIJ[0] * Vj
        d_Bwh[i4+2] += whj * DWIJ[1] * Vj
        d_Bwh[i4+3] += whj * DWIJ[2] * Vj

    def post_loop(self, d_idx, d_A, d_uho, d_Buh, d_vho, d_Bvh, d_who, d_Bwh):
        a_mat = declare('matrix(16)')
        aug_mat = declare('matrix(20)')
        b_uh = declare('matrix(4)')
        res_uh = declare('matrix(4)')
        b_vh = declare('matrix(4)')
        res_vh = declare('matrix(4)')
        b_wh = declare('matrix(4)')
        res_wh = declare('matrix(4)')
        i, n, i16, i4 = declare('int', 4)
        i16 = 16*d_idx
        i4 = 4*d_idx

        for i in range(16):
            a_mat[i] = d_A[i16+i]
        for i in range(20):
            aug_mat[i] = 0.0
        for i in range(4):
            b_uh[i] = d_Buh[i4+i]
            res_uh[i] = 0.0
            b_vh[i] = d_Bvh[i4+i]
            res_vh[i] = 0.0
            b_wh[i] = d_Bwh[i4+i]
            res_wh[i] = 0.0

        n = self.dim + 1
        augmented_matrix(a_mat, b_uh, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_uh)
        for i in range(4):
            d_uho[i4+i] = res_uh[i]
        augmented_matrix(a_mat, b_vh, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_vh)
        for i in range(4):
            d_vho[i4+i] = res_vh[i]
        augmented_matrix(a_mat, b_wh, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_wh)
        for i in range(4):
            d_who[i4+i] = res_wh[i]


class ExtrapolateUhat(Equation):
    def initialize(self, d_idx, d_uhat, d_vhat, d_what):
        d_uhat[d_idx] = 0.0
        d_vhat[d_idx] = 0.0
        d_what[d_idx] = 0.0

    def loop(self, d_idx, d_uhat, d_uho, d_vhat, d_vho, d_what, d_who,
             d_disp, d_xn, d_yn, d_zn):

        delx = 2 * d_disp[d_idx] * d_xn[d_idx]
        dely = 2 * d_disp[d_idx] * d_yn[d_idx]
        delz = 2 * d_disp[d_idx] * d_zn[d_idx]
        d_uhat[d_idx] = -1. * (
             d_uho[4*d_idx+0]
             - delx * d_uho[4*d_idx+1]
             - dely * d_uho[4*d_idx+2]
             - delz * d_uho[4*d_idx+3]
                      )

        d_vhat[d_idx] = (
             d_vho[4*d_idx+0]
             - delx * d_vho[4*d_idx+1]
             - dely * d_vho[4*d_idx+2]
             - delz * d_vho[4*d_idx+3]
                      )

        d_what[d_idx] = (
             d_who[4*d_idx+0]
             - delx * d_who[4*d_idx+1]
             - dely * d_who[4*d_idx+2]
             - delz * d_who[4*d_idx+3]
                      )


class CopyUhatFromGhost(Equation):
    def initialize_pair(self,
                        d_idx, d_uhat, s_uhat, d_vhat, s_vhat, d_what, s_what):
        d_uhat[d_idx] = -1.0 * s_uhat[d_idx]

        d_vhat[d_idx] = s_vhat[d_idx]

        d_what[d_idx] = s_what[d_idx]


class EvaluateU(Equation):
    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(EvaluateU, self).__init__(dest, sources)

    def initialize(self, d_idx, d_uo, d_Bu, d_vo, d_Bv, d_wo, d_Bw):
        i = declare('int')

        for i in range(3):
            d_uo[4*d_idx+i] = 0.0
            d_Bu[4*d_idx+i] = 0.0
            d_vo[4*d_idx+i] = 0.0
            d_Bv[4*d_idx+i] = 0.0
            d_wo[4*d_idx+i] = 0.0
            d_Bw[4*d_idx+i] = 0.0

    def loop(self, d_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y, d_z, s_rho,
             s_m, s_idx, XIJ, DWIJ,
             WIJ, s_u, d_Bu, s_v, d_Bv, s_w, d_Bw):
        Vj = s_m[s_idx] / s_rho[s_idx]
        uj = s_u[s_idx]
        vj = s_v[s_idx]
        wj = s_w[s_idx]
        i4 = declare('int')
        i4 = 4*d_idx

        d_Bu[i4+0] += uj * WIJ * Vj
        d_Bu[i4+1] += uj * DWIJ[0] * Vj
        d_Bu[i4+2] += uj * DWIJ[1] * Vj
        d_Bu[i4+3] += uj * DWIJ[2] * Vj
        d_Bv[i4+0] += vj * WIJ * Vj
        d_Bv[i4+1] += vj * DWIJ[0] * Vj
        d_Bv[i4+2] += vj * DWIJ[1] * Vj
        d_Bv[i4+3] += vj * DWIJ[2] * Vj
        d_Bw[i4+0] += wj * WIJ * Vj
        d_Bw[i4+1] += wj * DWIJ[0] * Vj
        d_Bw[i4+2] += wj * DWIJ[1] * Vj
        d_Bw[i4+3] += wj * DWIJ[2] * Vj

    def post_loop(self, d_idx, d_A, d_uo, d_Bu, d_vo, d_Bv, d_wo, d_Bw):
        a_mat = declare('matrix(16)')
        aug_mat = declare('matrix(20)')
        b_u = declare('matrix(4)')
        res_u = declare('matrix(4)')
        b_v = declare('matrix(4)')
        res_v = declare('matrix(4)')
        b_w = declare('matrix(4)')
        res_w = declare('matrix(4)')
        i, n, i16, i4 = declare('int', 4)
        i16 = 16*d_idx
        i4 = 4*d_idx

        for i in range(16):
            a_mat[i] = d_A[i16+i]
        for i in range(20):
            aug_mat[i] = 0.0
        for i in range(4):
            b_u[i] = d_Bu[i4+i]
            res_u[i] = 0.0
            b_v[i] = d_Bv[i4+i]
            res_v[i] = 0.0
            b_w[i] = d_Bw[i4+i]
            res_w[i] = 0.0

        n = self.dim + 1
        augmented_matrix(a_mat, b_u, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_u)
        for i in range(4):
            d_uo[i4+i] = res_u[i]
        augmented_matrix(a_mat, b_v, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_v)
        for i in range(4):
            d_vo[i4+i] = res_v[i]
        augmented_matrix(a_mat, b_w, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_w)
        for i in range(4):
            d_wo[i4+i] = res_w[i]


class ExtrapolateU(Equation):
    def initialize(self, d_idx, d_u, d_v, d_w):
        d_u[d_idx] = 0.0
        d_v[d_idx] = 0.0
        d_w[d_idx] = 0.0

    def loop(self, d_idx, d_u, d_uo, d_v, d_vo, d_w, d_wo,
             d_disp, d_xn, d_yn, d_zn):

        delx = 2 * d_disp[d_idx] * d_xn[d_idx]
        dely = 2 * d_disp[d_idx] * d_yn[d_idx]
        delz = 2 * d_disp[d_idx] * d_zn[d_idx]
        d_u[d_idx] = -1. * (
             d_uo[4*d_idx+0]
             - delx * d_uo[4*d_idx+1]
             - dely * d_uo[4*d_idx+2]
             - delz * d_uo[4*d_idx+3]
                      )

        d_v[d_idx] = (
             d_vo[4*d_idx+0]
             - delx * d_vo[4*d_idx+1]
             - dely * d_vo[4*d_idx+2]
             - delz * d_vo[4*d_idx+3]
                      )

        d_w[d_idx] = (
             d_wo[4*d_idx+0]
             - delx * d_wo[4*d_idx+1]
             - dely * d_wo[4*d_idx+2]
             - delz * d_wo[4*d_idx+3]
                      )


class CopyUFromGhost(Equation):
    def initialize_pair(self,
                        d_idx, d_u, s_u, d_v, s_v, d_w, s_w):
        d_u[d_idx] = -1.0 * s_u[d_idx]

        d_v[d_idx] = s_v[d_idx]

        d_w[d_idx] = s_w[d_idx]


class EvaluateP(Equation):
    def _get_helpers_(self):
        return [gj_solve, augmented_matrix]

    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(EvaluateP, self).__init__(dest, sources)

    def initialize(self, d_idx, d_po, d_Bp):
        i = declare('int')

        for i in range(3):
            d_po[4*d_idx+i] = 0.0
            d_Bp[4*d_idx+i] = 0.0

    def loop(self, d_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y, d_z, s_rho,
             s_m, s_idx, XIJ, DWIJ,
             WIJ, s_p, d_Bp):
        Vj = s_m[s_idx] / s_rho[s_idx]
        pj = s_p[s_idx]
        i4 = declare('int')
        i4 = 4*d_idx

        d_Bp[i4+0] += pj * WIJ * Vj
        d_Bp[i4+1] += pj * DWIJ[0] * Vj
        d_Bp[i4+2] += pj * DWIJ[1] * Vj
        d_Bp[i4+3] += pj * DWIJ[2] * Vj

    def post_loop(self, d_idx, d_A, d_po, d_Bp):
        a_mat = declare('matrix(16)')
        aug_mat = declare('matrix(20)')
        b_p = declare('matrix(4)')
        res_p = declare('matrix(4)')
        i, n, i16, i4 = declare('int', 4)
        i16 = 16*d_idx
        i4 = 4*d_idx

        for i in range(16):
            a_mat[i] = d_A[i16+i]
        for i in range(20):
            aug_mat[i] = 0.0
        for i in range(4):
            b_p[i] = d_Bp[i4+i]
            res_p[i] = 0.0

        n = self.dim + 1
        augmented_matrix(a_mat, b_p, n, 1, 4, aug_mat)
        gj_solve(aug_mat, n, 1, res_p)
        for i in range(4):
            d_po[i4+i] = res_p[i]


class ExtrapolateP(Equation):
    def initialize(self, d_idx, d_p):
        d_p[d_idx] = 0.0

    def loop(self, d_idx, d_p, d_po,
             d_disp, d_xn, d_yn, d_zn):

        delx = 2 * d_disp[d_idx] * d_xn[d_idx]
        dely = 2 * d_disp[d_idx] * d_yn[d_idx]
        delz = 2 * d_disp[d_idx] * d_zn[d_idx]
        d_p[d_idx] = (
             d_po[4*d_idx+0]
             - delx * d_po[4*d_idx+1]
             - dely * d_po[4*d_idx+2]
             - delz * d_po[4*d_idx+3]
                      )


class CopyPFromGhost(Equation):
    def initialize_pair(self,
                        d_idx, d_p, s_p):
        d_p[d_idx] = s_p[d_idx]


class UpdateMomentMatrix(Equation):
    def __init__(self, dest, sources, dim=1):
        self.dim = dim

        super(UpdateMomentMatrix, self).__init__(dest, sources)

    def initialize(self, d_idx, d_A):
        i, j = declare('int', 2)

        for i in range(4):
            for j in range(4):
                d_A[16*d_idx + j+4*i] = 0.0

    def loop(self, d_idx, s_idx, d_h, s_h, s_x, s_y, s_z, d_x, d_y,
             d_z, s_rho, s_m, d_A, XIJ, WIJ, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        i16 = declare('int')
        i16 = 16*d_idx
        d_A[i16+0] += WIJ * Vj

        d_A[i16+1] += -XIJ[0] * WIJ * Vj
        d_A[i16+2] += -XIJ[1] * WIJ * Vj
        d_A[i16+3] += -XIJ[2] * WIJ * Vj

        d_A[i16+4] += DWIJ[0] * Vj
        d_A[i16+8] += DWIJ[1] * Vj
        d_A[i16+12] += DWIJ[2] * Vj

        d_A[i16+5] += -XIJ[0] * DWIJ[0] * Vj
        d_A[i16+6] += -XIJ[1] * DWIJ[0] * Vj
        d_A[i16+7] += -XIJ[2] * DWIJ[0] * Vj

        d_A[i16+9] += -XIJ[0] * DWIJ[1] * Vj
        d_A[i16+10] += -XIJ[1] * DWIJ[1] * Vj
        d_A[i16+11] += -XIJ[2] * DWIJ[1] * Vj

        d_A[i16+13] += -XIJ[0] * DWIJ[2] * Vj
        d_A[i16+14] += -XIJ[1] * DWIJ[2] * Vj
        d_A[i16+15] += -XIJ[2] * DWIJ[2] * Vj



from compyle.api import declare


def identity(a=[0.0, 0.0], n=3):
    """Initialize an identity matrix.
    """
    i, j = declare('int', 2)
    for i in range(n):
        for j in range(n):
            if i == j:
                a[n*i + j] = 1.0
            else:
                a[n*i + j] = 0.0


def dot(a=[0.0, 0.0], b=[0.0, 0.0], n=3):
    i = declare('int')
    result = 0.0
    for i in range(n):
        result += a[i]*b[i]
    return result


def mat_mult(a=[1.0, 0.0], b=[1.0, 0.0], n=3, result=[0.0, 0.0]):
    """Multiply two square matrices (not element-wise).

    Stores the result in `result`.

    Parameters
    ----------

    a: list
    b: list
    n: int : number of rows/columns
    result: list
    """
    i, j, k = declare('int', 3)
    for i in range(n):
        for k in range(n):
            s = 0.0
            for j in range(n):
                s += a[n*i + j] * b[n*j + k]
            result[n*i + k] = s


def mat_vec_mult(a=[1.0, 0.0], b=[1.0, 0.0], n=3, result=[0.0, 0.0]):
    """Multiply a square matrix with a vector.

    Parameters
    ----------

    a: list
    b: list
    n: int : number of rows/columns
    result: list
    """
    i, j = declare('int', 2)
    for i in range(n):
        s = 0.0
        for j in range(n):
            s += a[n*i + j] * b[j]
        result[i] = s


def augmented_matrix(A=[0.0, 0.0], b=[0.0, 0.0], n=3, na=1, nmax=3,
                     result=[0.0, 0.0]):
    """Create augmented matrix.

    Given flattened matrix, `A` of max rows/columns `nmax`, and flattened
    columns `b` with `n` rows of interest and `na` additional columns, put
    these in `result`. Result must be already allocated and be flattened.
    The `result` will contain `(n + na)*n` first entries as the
    augmented_matrix.


    Parameters
    ----------
    A: list: given matrix.
    b: list: additional columns to be augmented.
    n: int : number of rows/columns to use from `A`.
    na: int: number of added columns in `b`.
    nmax: int: the maximum dimension 'A'
    result: list: must have size of (nmax + na)*n.
    """
    i, j, nt = declare('int', 3)
    nt = n + na
    for i in range(n):
        for j in range(n):
            result[nt*i + j] = A[nmax * i + j]
        for j in range(na):
            result[nt*i + n + j] = b[na*i + j]


def gj_solve(m=[1., 0.], n=3, nb=1, result=[0.0, 0.0]):
    r"""A gauss-jordan method to solve an augmented matrix.

    The routine is given the augmented matrix, the number of rows/cols in the
    original matrix and the number of added columns. The result is stored in
    the result array passed.

    Parameters
    ----------

    m : list: a flattened list representing the augmented matrix [A|b].
    n : int: number of columns/rows used from A in augmented_matrix.
    nb: int: number of columns added to A.
    result: list: with size n*nb

    References
    ----------
    https://ricardianambivalence.com/2012/10/20/pure-python-gauss-jordan
    -solve-ax-b-invert-a/

    """

    i, j, eqns, colrange, augCol, col, row, bigrow, nt = declare('int', 9)
    eqns = n
    colrange = n
    augCol = n + nb
    nt = n + nb

    for col in range(colrange):
        bigrow = col
        for row in range(col + 1, colrange):
            if abs(m[nt*row + col]) > abs(m[nt*bigrow + col]):
                bigrow = row
                temp = m[nt*row + col]
                m[nt*row + col] = m[nt*bigrow + col]
                m[nt*bigrow + col] = temp

    rr, rrcol, rb, rbr, kup, kupr, kleft, kleftr = declare('int', 8)
    for rrcol in range(0, colrange):
        for rr in range(rrcol + 1, eqns):
            dnr = float(m[nt*rrcol + rrcol])
            if abs(dnr) < 1e-12:
                return 1.0
            cc = -float(m[nt*rr + rrcol]) / dnr
            for j in range(augCol):
                m[nt*rr + j] = m[nt*rr + j] + cc * m[nt*rrcol + j]

    backCol, backColr = declare('int', 2)
    tol = 1.0e-12
    for rbr in range(eqns):
        rb = eqns - rbr - 1
        if (m[nt*rb + rb] == 0):
            if abs(m[nt*rb + augCol - 1]) > tol:
                # Error, singular matrix.
                return 1.0
        else:
            for backColr in range(rb, augCol):
                backCol = rb + augCol - backColr - 1
                m[nt*rb + backCol] = m[nt*rb + backCol] / m[nt*rb + rb]
            if not (rb == 0):
                for kupr in range(rb):
                    kup = rb - kupr - 1
                    for kleftr in range(rb, augCol):
                        kleft = rb + augCol - kleftr - 1
                        kk = -m[nt*kup + rb] / m[nt*rb + rb]
                        m[nt*kup + kleft] = (m[nt*kup + kleft] +
                                             kk * m[nt*rb + kleft])

    for i in range(n):
        for j in range(nb):
            result[nb*i + j] = m[nt*i + n + j]

    return 0.0



"""Tests for the simple_inlet_outlet.

Copyright (c) 2015, Prabhu Ramachandran
License: BSD
"""
import numpy as np
try:
    # This is for Python-2.6.x
    import unittest2 as unittest
except ImportError:
    import unittest

from pysph.base.utils import get_particle_array
from pysph.base.kernels import QuinticSpline
from pysph.sph.bc.inlet_outlet_manager import (
    InletInfo, OutletInfo, InletBase, OutletBase)


class TestSimpleInlet1D(unittest.TestCase):
    def setUp(self):
        dx = 0.1
        x = np.arange(-5*dx, 0.0, dx)
        m = np.ones_like(x)
        h = np.ones_like(x)*dx*1.5
        p = np.ones_like(x)*5.0
        self.inlet_pa = get_particle_array(name='inlet', x=x, m=m, h=h, p=p)
        # Empty particle array.
        self.dest_pa = get_particle_array(name='fluid')
        props = ['ioid', 'disp']
        for p in props:
            for pa_arr in [self.dest_pa, self.inlet_pa]:
                pa_arr.add_property(p)
        self.dx = dx
        self.kernel = QuinticSpline(dim=1)

        self.inletinfo = InletInfo('inlet', normal=[-1., 0., 0.],
                                   refpoint=[-dx/2, 0., 0.])
        self.inletinfo.length = 0.5

    def test_update_creates_particles_in_destination(self):
        # Given
        inlet = InletBase(
            self.inlet_pa, self.dest_pa, self.inletinfo,
            dim=1, kernel=self.kernel)
        # Two rows of particles should move out.
        self.inlet_pa.x += 0.12

        # When
        inlet.update(time=0.0, dt=0.0, stage=1)

        # Then
        x = self.inlet_pa.x
        p = self.inlet_pa.p
        h = self.inlet_pa.h
        self.assertEqual(len(x), 5)
        x_expect = (-np.arange(5, 0, -1)*self.dx + 0.12)
        x_expect[x_expect > 0.0] -= 0.5
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

        # The destination particle array should now have one particles.
        x = self.dest_pa.x
        p = self.dest_pa.p
        h = self.dest_pa.h
        self.assertEqual(self.dest_pa.get_number_of_particles(), 1)
        x_expect = (-np.arange(1, 0, -1)*self.dx + 0.12)
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

    def test_particles_should_update_in_given_stage(self):
        # Given
        inlet = InletBase(
            self.inlet_pa, self.dest_pa, self.inletinfo,
            dim=1, kernel=self.kernel)
        # Two rows of particles should move out.
        self.inlet_pa.x += 0.15
        inlet.active_stages = [1]

        # When
        inlet.update(time=0.0, dt=0.0, stage=2)

        # Then
        x = self.inlet_pa.x
        p = self.inlet_pa.p
        h = self.inlet_pa.h
        self.assertEqual(len(x), 5)
        x_expect = (-np.arange(5, 0, -1)*self.dx + 0.15)
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

        # The destination particle array should not have particles.
        self.assertEqual(self.dest_pa.get_number_of_particles(), 0)

    def test_inlet_calls_callback(self):
        # Given
        calls = []

        def _callback(d, i):
            calls.append((d, i))

        inlet = InletBase(
            self.inlet_pa, self.dest_pa, self.inletinfo,
            dim=1.0, kernel=self.kernel, callback=_callback)

        # When
        self.inlet_pa.x += 0.5
        inlet.update(time=0.0, dt=0.0, stage=1)

        # Then
        self.assertEqual(len(calls), 1)
        d_pa, i_pa = calls[0]
        self.assertEqual(d_pa, self.dest_pa)
        self.assertEqual(i_pa, self.inlet_pa)


class TestSimpleOutlet1D(unittest.TestCase):
    def setUp(self):
        dx = 0.1
        x = np.arange(-5*dx, 0.0, dx)
        m = np.ones_like(x)
        h = np.ones_like(x)*dx*1.5
        p = np.ones_like(x)*5.0
        self.source_pa = get_particle_array(name='fluid', x=x, m=m, h=h, p=p)
        # Empty particle array.
        self.outlet_pa = get_particle_array(name='outlet')
        props = ['ioid', 'disp']
        for p in props:
            for pa_arr in [self.source_pa, self.outlet_pa]:
                pa_arr.add_property(p)
        self.dx = dx
        self.kernel = QuinticSpline(dim=1)

        self.outletinfo = OutletInfo(
            'outlet', normal=[1., 0., 0.], refpoint=[-dx/2, 0., 0.],
            props_to_copy=self.source_pa.get_lb_props())
        self.outletinfo.length = 0.5

    def test_outlet_absorbs_particles_from_source(self):
        # Given
        outlet = OutletBase(
            self.outlet_pa, self.source_pa, self.outletinfo,
            dim=1, kernel=self.kernel)
        # Two rows of particles should move out.
        self.source_pa.x += 0.12

        # When
        outlet.update(time=0.0, dt=0.0, stage=1)

        # Then
        x = self.source_pa.x
        p = self.source_pa.p
        h = self.source_pa.h
        print(x)
        self.assertEqual(len(x), 4)
        x_expect = (-np.arange(5, 1, -1)*self.dx + 0.12)
        x_expect[x_expect > 0.0] -= 0.5
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

        # The outlet particle array should now have one particles.
        x = self.outlet_pa.x
        p = self.outlet_pa.p
        h = self.outlet_pa.h
        self.assertEqual(self.outlet_pa.get_number_of_particles(), 1)
        x_expect = (-np.arange(1, 0, -1)*self.dx + 0.12)
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

    def test_particles_should_update_in_given_stage(self):
        # Given
        outlet = OutletBase(
            self.outlet_pa, self.source_pa, self.outletinfo,
            dim=1, kernel=self.kernel)
        # Two rows of particles should move out.
        self.source_pa.x += 0.15
        outlet.active_stages = [1]

        # When
        outlet.update(time=0.0, dt=0.0, stage=2)

        # Then
        x = self.source_pa.x
        p = self.source_pa.p
        h = self.source_pa.h
        self.assertEqual(len(x), 5)
        x_expect = (-np.arange(5, 0, -1)*self.dx + 0.15)
        self.assertTrue(np.allclose(list(x), list(x_expect)))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(
            np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14)
        )

        # The outlet particle array should not have particles.
        self.assertEqual(self.outlet_pa.get_number_of_particles(), 0)

    def test_outlet_deletes_particles(self):
        # Given
        outlet = OutletBase(
            self.outlet_pa, self.source_pa, self.outletinfo,
            dim=1, kernel=self.kernel)
        # Two rows of particles should move out.
        self.source_pa.x += 0.5

        # When
        outlet.update(time=0.0, dt=0.0, stage=1)

        # Then
        self.assertEqual(self.source_pa.get_number_of_particles(), 0)
        self.assertEqual(self.outlet_pa.get_number_of_particles(), 5)

        # When
        self.outlet_pa.x += 0.12
        outlet.update(time=0.0, dt=0.0, stage=1)

        # The outlet particle array should delete one particle.
        self.assertEqual(self.outlet_pa.get_number_of_particles(), 4)

    def test_outlet_calls_callback(self):
        # Given
        calls = []

        def _callback(s, o):
            calls.append((s, o))

        outlet = OutletBase(
            self.outlet_pa, self.source_pa, self.outletinfo,
            dim=1.0, kernel=self.kernel, callback=_callback)

        # When
        self.source_pa.x += 0.5
        outlet.update(time=0.0, dt=0.0, stage=1)

        # Then
        self.assertEqual(len(calls), 1)
        s_pa, o_pa = calls[0]
        self.assertEqual(o_pa, self.outlet_pa)
        self.assertEqual(s_pa, self.source_pa)


if __name__ == '__main__':
    unittest.main()



"""Utility functions to interact with SPHysics particle data"""

from os.path import basename
import numpy
from pysph.base.utils import get_particle_array_wcsph as gpa

# Post-process module for VTK output

def sphysics2pysph(partfile, indat='INDAT', dim=3, vtk=True):
    """Load an SPHysics part file and input data

    Parameters:

    partfile : str
        SPHysics part file (eq IPART, PART_00032, etc)

    indat : str
        SPHysics input data file

    dim : int
        Dimension for SPHysics files

    vtk : bint
        Flag to dump VTK output

    Notes:

    The dimension is very important as the SPHysics particle data is
    different in the 2D and 3D cases.

    """
    data = numpy.loadtxt(partfile)

    # sanity check on the input file and problem dimension
    ncols = data.shape[-1]
    if ( (ncols == 9) and (dim == 2) ):
        raise RuntimeError('Possiblly inconsistent dim and SPHysics part file')

    input_data = numpy.loadtxt(indat)

    partbase = basename(partfile)

    if partbase.startswith('IPART'):
        fileno = 0
    else:
        fileno = int( partbase.split('_')[-1] )

    # number of fluid and total number of particles. This is very
    # dangerous and relies on the SPHysics manual (pg. 38)
    dx = float( input_data[21] )
    dy = float( input_data[22] )
    dz = float( input_data[23] )
    h = float( input_data[24] )

    np = int(input_data[25])
    nb = int(input_data[26])
    nbf = int(input_data[27])

    # now load the individual arrays
    if dim == 3:
        x = data[:, 0]; y = data[:, 1]; z = data[:, 2]
        u = data[:, 3]; v = data[:, 4]; w = data[:, 5]

        rho = data[:, 6]; p = data[:, 7]; m = data[:, 8]

    else:
        x = data[:, 0]; z = data[:, 1]
        u = data[:, 2]; w = data[:, 3]

        rho = data[:, 4]; p = data[:, 5]; m = data[:, 6]

    # smoothing lengths
    h = numpy.ones_like(x) * h

    # now create the PySPH arrays
    fluid = gpa(
        name='fluid', x=x[nb:], y=y[nb:], z=z[nb:], u=u[nb:],
        v=v[nb:], w=w[nb:], rho=rho[nb:], p=p[nb:], m=m[nb:],
        h=h[nb:])

    solid = gpa(
        name='boundary', x=x[:nb], y=y[:nb], z=z[:nb], u=u[:nb],
        v=v[:nb], w=w[:nb], rho=rho[:nb], p=p[:nb], m=m[:nb],
        h=h[:nb])

    # PySPH arrays
    arrays = [fluid, solid]

    # Dump out vtk files for Paraview viewing
    if vtk:
        from .pprocess import PySPH2VTK
        props = ['u', 'v', 'w', 'rho', 'p', 'vmag', 'tag']
        pysph2vtk = PySPH2VTK(arrays, fileno=fileno)

        pysph2vtk.write_vtk('fluid', props)
        pysph2vtk.write_vtk('boundary', props)

    # return the list of arrays
    return arrays



"""Basic SPH Equations
###################


References
----------

.. [Monaghan1992] J. Monaghan, Smoothed Particle Hydrodynamics, "Annual Review
    of Astronomy and Astrophysics", 30 (1992), pp. 543-574.

.. [Monaghan2005] J. Monaghan, "Smoothed particle hydrodynamics", Reports on
    Progress in Physics, 68 (2005), pp. 1703-1759.

"""

from pysph.sph.equation import Equation


class SummationDensity(Equation):
    r"""Good old Summation density:

    :math:`\rho_a = \sum_b m_b W_{ab}`

    """
    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_rho, s_idx, s_m, WIJ):
        d_rho[d_idx] += s_m[s_idx]*WIJ


class BodyForce(Equation):
    r"""Add a body force to the particles:

    :math:`\boldsymbol{f} = f_x, f_y, f_z`

    """
    def __init__(self, dest, sources, fx=0.0, fy=0.0, fz=0.0):

        r"""
        Parameters
        ----------
        fx : float
            Body force per unit mass along the x-axis
        fy : float
            Body force per unit mass along the y-axis
        fz : float
            Body force per unit mass along the z-axis
        """

        self.fx = fx
        self.fy = fy
        self.fz = fz
        super(BodyForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] += self.fx
        d_av[d_idx] += self.fy
        d_aw[d_idx] += self.fz


class VelocityGradient2D(Equation):
    r""" Compute the SPH evaluation for the velocity gradient tensor in 2D.

    The expression for the velocity gradient is:

    :math:`\frac{\partial v^i}{\partial x^j} = \sum_{b}\frac{m_b}{\rho_b}(v_b
    - v_a)\frac{\partial W_{ab}}{\partial x_a^j}`

    Notes
    -----
    The tensor properties are stored in the variables v_ij where 'i'
    refers to the velocity component and 'j' refers to the spatial
    component. Thus v_10 is :math:`\frac{\partial v}{\partial x}`

    """
    def initialize(self, d_idx, d_v00, d_v01, d_v10, d_v11):
        d_v00[d_idx] = 0.0
        d_v01[d_idx] = 0.0
        d_v10[d_idx] = 0.0
        d_v11[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, s_rho,
             d_v00, d_v01, d_v10, d_v11,
             DWIJ, VIJ):

        tmp = s_m[s_idx]/s_rho[s_idx]

        d_v00[d_idx] += tmp * -VIJ[0] * DWIJ[0]
        d_v01[d_idx] += tmp * -VIJ[0] * DWIJ[1]

        d_v10[d_idx] += tmp * -VIJ[1] * DWIJ[0]
        d_v11[d_idx] += tmp * -VIJ[1] * DWIJ[1]


class VelocityGradient3D(Equation):
    r""" Compute the SPH evaluation for the velocity gradient tensor in 2D.

    The expression for the velocity gradient is:

    :math:`\frac{\partial v^i}{\partial x^j} = \sum_{b}\frac{m_b}{\rho_b}(v_b
    - v_a)\frac{\partial W_{ab}}{\partial x_a^j}`

    Notes
    -----
    The tensor properties are stored in the variables v_ij where 'i'
    refers to the velocity component and 'j' refers to the spatial
    component. Thus v_21 is :math:`\frac{\partial v}{\partial x}`

    """
    def initialize(self, d_idx, d_v00, d_v01, d_v02, d_v10, d_v11, d_v12,
                   d_v20, d_v21, d_v22):
        d_v00[d_idx] = 0.0
        d_v01[d_idx] = 0.0
        d_v02[d_idx] = 0.0

        d_v10[d_idx] = 0.0
        d_v11[d_idx] = 0.0
        d_v12[d_idx] = 0.0

        d_v20[d_idx] = 0.0
        d_v21[d_idx] = 0.0
        d_v22[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, s_rho,
             d_v00, d_v01, d_v02,
             d_v10, d_v11, d_v12,
             d_v20, d_v21, d_v22,
             DWIJ, VIJ):

        tmp = s_m[s_idx]/s_rho[s_idx]

        d_v00[d_idx] += tmp * -VIJ[0] * DWIJ[0]
        d_v01[d_idx] += tmp * -VIJ[0] * DWIJ[1]
        d_v02[d_idx] += tmp * -VIJ[0] * DWIJ[2]

        d_v10[d_idx] += tmp * -VIJ[1] * DWIJ[0]
        d_v11[d_idx] += tmp * -VIJ[1] * DWIJ[1]
        d_v12[d_idx] += tmp * -VIJ[1] * DWIJ[2]

        d_v20[d_idx] += tmp * -VIJ[2] * DWIJ[0]
        d_v21[d_idx] += tmp * -VIJ[2] * DWIJ[1]
        d_v22[d_idx] += tmp * -VIJ[2] * DWIJ[2]


class IsothermalEOS(Equation):
    r""" Compute the pressure using the Isothermal equation of state:

    :math:`p = p_0 + c_0^2(\rho_0 - \rho)`

    """
    def __init__(self, dest, sources, rho0, c0, p0):

        r"""
        Parameters
        ----------
        rho0 : float
            Reference density of the fluid (:math:`\rho_0`)
        c0 : float
            Maximum speed of sound expected in the system (:math:`c0`)
        p0 : float
            Reference pressure in the system (:math:`p0`)
        """

        self.rho0 = rho0
        self.c0 = c0
        self.c02 = c0 * c0
        self.p0 = p0
        super(IsothermalEOS, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, d_p):
        d_p[d_idx] = self.p0 + self.c02 * (d_rho[d_idx] - self.rho0)


class ContinuityEquation(Equation):
    r"""Density rate:

    :math:`\frac{d\rho_a}{dt} = \sum_b m_b \boldsymbol{v}_{ab}\cdot
    \nabla_a W_{ab}`

    """
    def initialize(self, d_idx, d_arho):
        d_arho[d_idx] = 0.0

    def loop(self, d_idx, d_arho, s_idx, s_m, DWIJ, VIJ):
        vijdotdwij = DWIJ[0]*VIJ[0] + DWIJ[1]*VIJ[1] + DWIJ[2]*VIJ[2]
        d_arho[d_idx] += s_m[s_idx]*vijdotdwij


class MonaghanArtificialViscosity(Equation):
    r"""Classical Monaghan style artificial viscosity [Monaghan2005]_

    .. math::

        \frac{d\mathbf{v}_{a}}{dt}&=&-\sum_{b}m_{b}\Pi_{ab}\nabla_{a}W_{ab}

    where

    .. math::

        \Pi_{ab}=\begin{cases}\frac{-\alpha_{\pi}\bar{c}_{ab}\phi_{ab}+
        \beta_{\pi}\phi_{ab}^{2}}{\bar{\rho}_{ab}}, & \mathbf{v}_{ab}\cdot
        \mathbf{r}_{ab}<0\\0, & \mathbf{v}_{ab}\cdot\mathbf{r}_{ab}\geq0
        \end{cases}

    with

    .. math::

        \phi_{ab}=\frac{h\mathbf{v}_{ab}\cdot\mathbf{r}_{ab}}
        {|\mathbf{r}_{ab}|^{2}+\epsilon^{2}}\\

        \bar{c}_{ab}&=&\frac{c_{a}+c_{b}}{2}\\

        \bar{\rho}_{ab}&=&\frac{\rho_{a}+\rho_{b}}{2}

    """
    def __init__(self, dest, sources, alpha=1.0, beta=1.0):
        r"""
        Parameters
        ----------
        alpha : float
            produces a shear and bulk viscosity
        beta : float
            used to handle high Mach number shocks
        """
        self.alpha = alpha
        self.beta = beta
        super(MonaghanArtificialViscosity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, d_cs, d_au, d_av, d_aw, s_m,
             s_rho, s_cs, VIJ, XIJ, HIJ, R2IJ, RHOIJ1, EPS, DWIJ):

        vijdotxij = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]

        piij = 0.0
        if vijdotxij < 0:
            cij = 0.5 * (d_cs[d_idx] + s_cs[s_idx])

            muij = (HIJ * vijdotxij)/(R2IJ + EPS)

            piij = -self.alpha*cij*muij + self.beta*muij*muij
            piij = piij*RHOIJ1

        d_au[d_idx] += -s_m[s_idx] * piij * DWIJ[0]
        d_av[d_idx] += -s_m[s_idx] * piij * DWIJ[1]
        d_aw[d_idx] += -s_m[s_idx] * piij * DWIJ[2]


class XSPHCorrection(Equation):
    r"""Position stepping with XSPH correction [Monaghan1992]_

    .. math::

        \frac{d\mathbf{r}_{a}}{dt}=\mathbf{\hat{v}}_{a}=\mathbf{v}_{a}-
        \epsilon\sum_{b}m_{b}\frac{\mathbf{v}_{ab}}{\bar{\rho}_{ab}}W_{ab}

    """
    def __init__(self, dest, sources, eps=0.5):
        r"""
        Parameters
        ----------
        eps : float
            :math:`\epsilon` as in the above equation

        Notes
        -----
        This equation must be used to advect the particles. XSPH can be
        turned off by setting the parameter ``eps = 0``.
        """

        self.eps = eps
        super(XSPHCorrection, self).__init__(dest, sources)

    def initialize(self, d_idx, d_ax, d_ay, d_az):
        d_ax[d_idx] = 0.0
        d_ay[d_idx] = 0.0
        d_az[d_idx] = 0.0

    def loop(self, s_idx, d_idx, s_m, d_ax, d_ay, d_az, WIJ, RHOIJ1, VIJ):
        tmp = -self.eps * s_m[s_idx]*WIJ*RHOIJ1

        d_ax[d_idx] += tmp * VIJ[0]
        d_ay[d_idx] += tmp * VIJ[1]
        d_az[d_idx] += tmp * VIJ[2]

    def post_loop(self, d_idx, d_ax, d_ay, d_az, d_u, d_v, d_w):
        d_ax[d_idx] += d_u[d_idx]
        d_ay[d_idx] += d_v[d_idx]
        d_az[d_idx] += d_w[d_idx]


class XSPHCorrectionForLeapFrog(Equation):
    r"""The XSPH correction [Monaghan1992]_ alone.  This is meant to be used
    with a leap-frog integrator which already considers the velocity of the
    particles.  It simply computes the correction term and adds that to ``ax,
    ay, az``.

    .. math::

        \frac{d\mathbf{r}_{a}}{dt}=\mathbf{\hat{v}}_{a}= -
        \epsilon\sum_{b}m_{b}\frac{\mathbf{v}_{ab}}{\bar{\rho}_{ab}}W_{ab}

    """
    def __init__(self, dest, sources, eps=0.5):
        r"""
        Parameters
        ----------
        eps : float
            :math:`\epsilon` as in the above equation

        Notes
        -----
        This equation must be used to advect the particles. XSPH can be
        turned off by setting the parameter ``eps = 0``.
        """

        self.eps = eps
        super(XSPHCorrectionForLeapFrog, self).__init__(dest, sources)

    def initialize(self, d_idx, d_ax, d_ay, d_az):
        d_ax[d_idx] = 0.0
        d_ay[d_idx] = 0.0
        d_az[d_idx] = 0.0

    def loop(self, s_idx, d_idx, s_m, d_ax, d_ay, d_az, WIJ, RHOIJ1, VIJ):
        tmp = -self.eps * s_m[s_idx]*WIJ*RHOIJ1

        d_ax[d_idx] += tmp * VIJ[0]
        d_ay[d_idx] += tmp * VIJ[1]
        d_az[d_idx] += tmp * VIJ[2]



"""Dam break flow against a tall structure

The case is based of an experiment done by Yeh and Petroff,
University of Washington.
Refer to "Peter E. Raad, Razvan Bidoae, The three-dimensional
Eulerian-Lagrangian marker and micro cell method
for the simulation of free surface flows, Journal of Computational Physics,
Volume 203, Issue 2, 2005, Pages 668-699,
ISSN 0021-9991"

DOI: https://doi.org/10.1016/j.jcp.2004.09.013

"""

import numpy as np

from pysph.base.utils import get_particle_array
from pysph.tools.geometry import remove_overlap_particles
from pysph.base.kernels import CubicSpline
from pysph.examples._db_geometry import DamBreak3DGeometry
from pysph.solver.application import Application
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.scheme import WCSPHScheme

dim = 3

tf = 1.4

H = 0.3
dx = H/25
nboundary_layers = 2
hdx = 1.32
ro = 1000.0
h0 = dx * hdx
gamma = 7.0
alpha = 0.02
beta = 0.0

c0 = 16.0 * np.sqrt(9.81 * H)


class DamBreak3D(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx',  default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx',  default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        dx = self.options.dx
        self.dx = dx
        self.hdx = self.options.hdx
        self.h0 = self.hdx * self.dx
        self.geom = DamBreak3DGeometry(
            container_height=0.4, container_width=0.61, container_length=1.6,
            fluid_column_height=0.3, fluid_column_width=0.61,
            fluid_column_length=0.4, obstacle_center_x=0.96,
            obstacle_center_y=0, obstacle_length=0.12,
            obstacle_height=0.75, obstacle_width=0.12,
            nboundary_layers=nboundary_layers, with_obstacle=True,
            dx=dx, hdx=hdx, rho0=ro
        )
        self.co = c0

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], ['boundary', 'obstacle'], dim=dim, rho0=ro, c0=c0,
            h0=h0, hdx=hdx, gz=-9.81, alpha=alpha, beta=beta, gamma=gamma,
            hg_correction=True, delta_sph=True, delta=0.1
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        hdx = self.hdx
        kernel = CubicSpline(dim=dim)
        h0 = self.dx * hdx
        s.configure(h0=h0, hdx=hdx)
        dt = 0.125 * h0 / c0
        s.configure_solver(
            kernel=kernel, integrator_cls=EPECIntegrator, tf=tf, dt=dt,
            adaptive_timestep=True
        )

    def create_particles(self):
        dx = self.dx
        [fluid, boundary, obstacle] = self.geom.create_particles()

        # Thin sheet of water of height 1 cm
        x_sheet, y_sheet, z_sheet = np.mgrid[0.4:1.6:dx, -0.31:0.31:dx, dx:dx+0.01:dx]
        x_sheet = x_sheet.ravel()
        y_sheet = y_sheet.ravel()
        z_sheet = z_sheet.ravel()
        sheet = get_particle_array(x=x_sheet, y=y_sheet, z=z_sheet,
                                   h=self.h0, rho=ro, m=ro*dx*dx*dx)

        remove_overlap_particles(sheet, obstacle, dx_solid=dx)
        remove_overlap_particles(sheet, boundary, dx_solid=dx)
        remove_overlap_particles(sheet, fluid, dx_solid=dx)
        fluid.append_parray(sheet)

        particles = [fluid, boundary, obstacle]
        self.scheme.setup_properties(particles)
        return particles

    def customize_output(self):
        self._mayavi_config('''
        viewer.scalar = 'u'
        b = particle_arrays['boundary']
        b.plot.actor.mapper.scalar_visibility = False
        b.plot.actor.property.opacity = 0.1
        ''')

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        import matplotlib.pyplot as plt
        from pysph.examples import db_exp_data as dbd
        from pysph.tools.interpolator import Interpolator
        import os

        exp_vt, exp_v, exp_ft, exp_f = dbd.get_yeh_petroff_data()
        files = self.output_files

        t = []
        u = []

        # Velocity probe location
        vp_x = 0.814
        vp_y = 0.0
        vp_z = 0.026

        for sd, arrays in iter_output(files, "fluid"):
            t.append(sd["t"])
            interp = Interpolator([arrays],
                                  x=vp_x, y=vp_y, z=vp_z, method="shepard")
            u.append(interp.interpolate('u'))

        u = np.array(u)
        fname = os.path.join(self.output_dir, 'results.npz')
        t, u = list(map(np.asarray, (t, u)))
        np.savez(fname, t=t, u=u)

        t = np.array(t)-0.238
        figure_1 = plt.figure()
        plt.plot(t, u, label="Computed", figure=figure_1)
        plt.scatter(exp_vt, exp_v, label="Experiment, Yeh and Petroff",
                    marker="^", figure=figure_1, color=(0, 0, 0))
        plt.legend()
        plt.ylabel("Horizontal Velocity (m/s)")
        plt.xlabel("Time (s)")
        left, right = plt.xlim()
        plt.xlim(left, 1.4)
        plt.savefig(os.path.join(self.output_dir, 'v_vs_t.png'))
        plt.show()


if __name__ == '__main__':
    app = DamBreak3D()
    app.run()
    app.post_process(app.info_filename)



# cython: language_level=3, embedsignature=True
# distutils: language=c++

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair

from cython.operator cimport dereference as deref, preincrement as inc

from .gpu_nnps_helper import GPUNNPSHelper

import pyopencl as cl
import pyopencl.array
import pyopencl.algorithm
from pyopencl.scan import GenericScanKernel
from pyopencl.elementwise import ElementwiseKernel

from compyle.array import Array
import compyle.array as array
from compyle.opencl import get_context

# Cython for compiler directives
cimport cython

import numpy as np
cimport numpy as np


IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

    @cython.cdivision(True)
    cdef inline double log2(double n) nogil:
        return log(n)/log(2)


cdef class StratifiedSFCGPUNNPS(GPUNNPS):
    def __init__(self, int dim, list particles, double radius_scale=2.0,
            int ghost_layers=1, domain=None, bint fixed_h=False,
            bint cache=True, bint sort_gids=False,
            int num_levels=2, backend='opencl'):
        GPUNNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids, backend
        )

        self.radius_scale2 = radius_scale*radius_scale

        self.helper = GPUNNPSHelper("stratified_sfc_gpu_nnps.mako",
                                    use_double=self.use_double,
                                    backend=self.backend)
        self.eps = 16*np.finfo(np.float32).eps

        self.num_levels = num_levels
        self.src_index = -1
        self.dst_index = -1
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    @cython.cdivision(True)
    cpdef _refresh(self):
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, j, num_particles
        self.pids = []
        self.pid_keys = []
        self.start_idx_levels = []
        self.num_particles_levels = []
        self._sorted = False

        if self.cell_size - self.hmin > self.hmin*self.eps:
            self.interval_size = \
                    (self.cell_size - self.hmin)*(1 + self.eps)/self.num_levels
        else:
            self.interval_size = self.hmin*self.eps

        for i from 0<=i<self.narrays:
            pa_wrapper = <NNPSParticleArrayWrapper>self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.pids.append(array.empty(num_particles, dtype=np.uint32,
                             backend=self.backend))
            self.pid_keys.append(array.empty(num_particles, dtype=np.uint64,
                                 backend=self.backend))
            start_idx_i = num_particles + array.zeros(self.num_levels,
                    dtype=np.uint32, backend=self.backend)
            self.start_idx_levels.append(start_idx_i)
            self.num_particles_levels.append(array.zeros_like(start_idx_i))

        cdef double max_length = fmax(fmax((self.xmax[0] - self.xmin[0]),
            (self.xmax[1] - self.xmin[1])), (self.xmax[2] - self.xmin[2]))

        cdef int max_num_cells = (<int> ceil(max_length/self.hmin))

        self.max_num_bits = 1 + 3*(<int> ceil(log2(max_num_cells)))

    cpdef get_spatially_ordered_indices(self, int pa_index):
        pass

    cpdef _bin(self, int pa_index):
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]

        fill_pids = self.helper.get_kernel("fill_pids")

        levels = array.empty(pa_wrapper.get_number_of_particles(),
                             dtype=np.int32, backend=self.backend)

        pa_gpu = pa_wrapper.pa.gpu
        fill_pids(pa_gpu.x.dev, pa_gpu.y.dev, pa_gpu.z.dev, pa_gpu.h.dev,
                self.interval_size, self.xmin[0], self.xmin[1], self.xmin[2],
                self.hmin, self.pid_keys[pa_index].dev, self.pids[pa_index].dev,
                self.radius_scale, self.max_num_bits)

        radix_sort = cl.algorithm.RadixSort(get_context(),
                "unsigned int* pids, unsigned long* keys",
                scan_kernel=GenericScanKernel, key_expr="keys[i]",
                sort_arg_names=["pids", "keys"])

        cdef int max_num_bits = <int> (self.max_num_bits + \
                ceil(log2(self.num_levels)))

        (sorted_indices, sorted_keys), evnt = radix_sort(self.pids[pa_index].dev,
                self.pid_keys[pa_index].dev, key_bits=max_num_bits)
        self.pids[pa_index].set_data(sorted_indices)
        self.pid_keys[pa_index].set_data(sorted_keys)

        #FIXME: This will only work on OpenCL and CUDA backends
        cdef unsigned long long key = <unsigned long long> (sorted_keys[0].get())

        self.start_idx_levels[pa_index][key >> self.max_num_bits] = 0

        fill_start_indices = self.helper.get_kernel("fill_start_indices")

        fill_start_indices(self.pid_keys[pa_index].dev,
                self.start_idx_levels[pa_index].dev,
                self.max_num_bits, self.num_particles_levels[pa_index].dev)

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        GPUNNPS.set_context(self, src_index, dst_index)

        self.src = self.pa_wrappers[src_index]
        self.dst = self.pa_wrappers[dst_index]


    cdef void find_neighbor_lengths(self, nbr_lengths):
        find_nbr_lengths = self.helper.get_kernel("find_nbr_lengths",
                sorted=self._sorted)

        make_vec = cl.cltypes.make_double3 if self.use_double \
                else cl.cltypes.make_float3

        mask_lengths = array.zeros(self.dst.get_number_of_particles(),
                dtype=np.int32, backend=self.backend)

        dst_gpu = self.dst.pa.gpu
        src_gpu = self.src.pa.gpu
        find_nbr_lengths(dst_gpu.x.dev, dst_gpu.y.dev, dst_gpu.z.dev,
                dst_gpu.h.dev, src_gpu.x.dev, src_gpu.y.dev, src_gpu.z.dev,
                src_gpu.h.dev,
                make_vec(self.xmin[0], self.xmin[1], self.xmin[2]),
                self.src.get_number_of_particles(),
                self.pid_keys[self.src_index].dev,
                self.pids[self.dst_index].dev, self.pids[self.src_index].dev,
                nbr_lengths.dev, self.radius_scale, self.hmin,
                self.interval_size, self.start_idx_levels[self.src_index].dev,
                self.max_num_bits, self.num_levels,
                self.num_particles_levels[self.src_index].dev)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices):
        find_nbrs = self.helper.get_kernel("find_nbrs",
                sorted=self._sorted)

        make_vec = cl.cltypes.make_double3 if self.use_double \
                else cl.cltypes.make_float3

        dst_gpu = self.dst.pa.gpu
        src_gpu = self.src.pa.gpu
        find_nbrs(dst_gpu.x.dev, dst_gpu.y.dev, dst_gpu.z.dev,
                dst_gpu.h.dev, src_gpu.x.dev, src_gpu.y.dev, src_gpu.z.dev,
                src_gpu.h.dev,
                make_vec(self.xmin[0], self.xmin[1], self.xmin[2]),
                self.src.get_number_of_particles(),
                self.pid_keys[self.src_index].dev,
                self.pids[self.dst_index].dev, self.pids[self.src_index].dev,
                start_indices.dev, nbrs.dev, self.radius_scale, self.hmin,
                self.interval_size, self.start_idx_levels[self.src_index].dev,
                self.max_num_bits, self.num_levels,
                self.num_particles_levels[self.src_index].dev)



"""Mirroring outlet

Original paper by

    - Tafuni, Angelantonio, et al. "A versatile algorithm for the treatment of
      open boundary conditions in smoothed particle hydrodynamics gpu models."
      Computer Methods in Applied Mechanics and Engineering 342 (2018):
      604-624.

The implementation here is the modification of this by

    - Negi, Pawan, Prabhu Ramachandran, and Asmelash Haftu. "An improved
      non-reflecting outlet boundary condition for weakly-compressible SPH."
      arXiv preprint arXiv:1907.04034 (2019).

"""

from pysph.sph.integrator import PECIntegrator
from pysph.sph.wc.edac import EDACScheme
from pysph.sph.bc.inlet_outlet_manager import InletOutletManager


class SimpleInletOutlet(InletOutletManager):
    def add_io_properties(self, pa, scheme=None):
        DEFAULT_PROPS = [
            'disp', 'ioid', 'Bu', 'Bv', 'Bw', 'Bp', 'xn', 'yn', 'zn', 'A',
            'wij', 'uo', 'vo', 'wo', 'po', 'uho', 'vho', 'who', 'Buh',
            'Bvh', 'Bwh', 'x0', 'y0', 'z0', 'uhat', 'vhat', 'what']
        STRIDE_DATA = {
            'A': 16, 'Bu': 4, 'Bv': 4, 'Bw': 4, 'Bp': 4, 'Brho': 4, 'uo':
            4, 'vo': 4, 'wo': 4, 'po': 4, 'rhoo': 4, 'Buh': 4, 'Bvh': 4,
            'Bwh': 4, 'uho': 4, 'vho': 4, 'who': 4}
        for prop in DEFAULT_PROPS:
            if prop in STRIDE_DATA:
                pa.add_property(prop, stride=STRIDE_DATA[prop])
            else:
                pa.add_property(prop)

        pa.add_constant('uref', 0.0)

    def get_stepper(self, scheme, cls, edactvf=True):
        from pysph.sph.bc.inlet_outlet_manager import (
            InletStep, OutletStepWithUhat)
        steppers = {}
        if (cls == PECIntegrator):
            if isinstance(scheme, EDACScheme):
                for inlet in self.inlets:
                    steppers[inlet] = InletStep()
                for outlet in self.outlets:
                    steppers[outlet] = OutletStepWithUhat()
                for g_inlet in self.ghost_inlets:
                    steppers[g_inlet] = InletStep()
                for g_outlet in self.ghost_outlets:
                    steppers[g_outlet] = OutletStepWithUhat()
                self.active_stages = [2]

        return steppers

    def get_equations(self, scheme=None, summation_density=False,
                      edactvf=True):
        from pysph.sph.equation import Group
        from pysph.sph.bc.interpolate import (
            UpdateMomentMatrix, EvaluateUhat, EvaluateP, EvaluateU,
            ExtrapolateUhat, ExtrapolateP, ExtrapolateU,
            CopyUhatFromGhost, CopyPFromGhost, CopyUFromGhost)
        from pysph.sph.bc.inlet_outlet_manager import (
            UpdateNormalsAndDisplacements, CopyNormalsandDistances)

        all_ghosts = self.ghost_inlets + self.ghost_outlets
        all_info = self.inletinfo + self.outletinfo
        all_pairs = {**self.inlet_pairs, **self.outlet_pairs}

        equations = []
        g00 = []
        i = 0
        for info in all_info:
            g00.append(UpdateNormalsAndDisplacements(
                dest=info.pa_name, sources=None, xn=info.normal[0],
                yn=info.normal[1], zn=info.normal[2], xo=info.refpoint[0],
                yo=info.refpoint[1], zo=info.refpoint[2]
                ))
            g00.append(CopyNormalsandDistances(
                dest=all_pairs[info.pa_name], sources=[info.pa_name]))
            i = i + 1

        equations.append(Group(equations=g00, real=False))

        g02 = []
        for name in all_ghosts:
            g02.append(UpdateMomentMatrix(
                dest=name, sources=self.fluids, dim=self.dim))

        equations.append(Group(equations=g02, real=False))

        g03 = []
        for name in all_ghosts:
            g03.append(EvaluateUhat(dest=name, sources=self.fluids,
                                    dim=self.dim))
            g03.append(EvaluateP(dest=name, sources=self.fluids,
                                 dim=self.dim))
        for name in self.ghost_outlets:
            g03.append(EvaluateU(dest=name, sources=self.fluids,
                                 dim=self.dim))

        equations.append(Group(equations=g03, real=False))

        g04 = []
        for name in all_ghosts:
            g04.append(ExtrapolateUhat(dest=name, sources=None))
            g04.append(ExtrapolateP(dest=name, sources=None))
        for name in self.ghost_outlets:
            g04.append(ExtrapolateU(dest=name, sources=None))

        equations.append(Group(equations=g04, real=False))

        g05 = []
        for io in all_pairs.keys():
            g05.append(CopyUhatFromGhost(
                dest=io, sources=[all_pairs[io]]))
            g05.append(CopyPFromGhost(
                dest=io, sources=[all_pairs[io]]))
        for io in self.outlet_pairs.keys():
            g05.append(CopyUFromGhost(
                dest=io, sources=[all_pairs[io]]))

        equations.append(Group(equations=g05, real=False))

        g06 = []
        for inlet in self.inletinfo:
            for eqn in inlet.equations:
                g06.append(eqn)
        for outlet in self.outletinfo:
            for eqn in outlet.equations:
                g06.append(eqn)

        equations.append(Group(equations=g06, real=False))

        return equations



""" The Kelvin-Helmholtz instability test (1.5 hours)
"""

# NumPy and standard library imports
import numpy

# PySPH base and carray imports
from pysph.base.utils import get_particle_array as gpa
from pysph.solver.application import Application
from pysph.sph.scheme import GasDScheme, SchemeChooser, ADKEScheme, GSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.base.nnps import DomainManager
from pysph.tools import uniform_distribution as ud


# problem constants
dim = 2
gamma = 5.0/3.0

xmin = ymin = 0.0
xmax = ymax = 1.0

rhoi_1 = 1
rhoi_2 = 2
rhoi_m = 0.5 * (rhoi_1 - rhoi_2)

v_i1 = 0.5
v_i2 = -0.5
v_im = 0.5 * (v_i1 - v_i2)

delta = 0.025
dely = 0.01
wavelen = 0.5

dt = 1e-3

tf = 2


class KHInstability(Application):
    def initialize(self):
        self.nx = 200
        self.dx = (xmax - xmin) / self.nx
        self.dy = self.dx
        self.hdx = 1.5

    def create_particles(self):
        data = ud.uniform_distribution_cubic2D(
            self.dx, xmin, xmax, ymin, ymax
        )

        x = data[0].ravel()
        y = data[1].ravel()

        y1 = numpy.where(
            (y >= 0) & (y < 0.25)
        )[0]

        y2 = numpy.where(
            (y >= 0.25) & (y < 0.5)
        )[0]

        y3 = numpy.where(
            (y >= 0.5) & (y < 0.75)
        )[0]

        y4 = numpy.where(
            (y >= 0.75) & (y < 1.0)
        )[0]

        rho1 = rhoi_1 - rhoi_m * numpy.exp((y[y1] - 0.25)/delta)
        rho2 = rhoi_2 + rhoi_m * numpy.exp((0.25 - y[y2])/delta)
        rho3 = rhoi_2 + rhoi_m * numpy.exp((y[y3] - 0.75)/delta)
        rho4 = rhoi_1 - rhoi_m * numpy.exp((0.75 - y[y4])/delta)

        u1 = v_i1 - v_im * numpy.exp((y[y1] - 0.25)/delta)
        u2 = v_i2 + v_im * numpy.exp((0.25 - y[y2])/delta)
        u3 = v_i2 + v_im * numpy.exp((y[y3] - 0.75)/delta)
        u4 = v_i1 - v_im * numpy.exp((0.75 - y[y4])/delta)

        v = dely * numpy.sin(
            2 * numpy.pi * x / wavelen
        )

        p = 2.5

        rho = numpy.concatenate((
            rho1, rho2, rho3, rho4
        ))

        u = numpy.concatenate((
            u1, u2, u3, u4
        ))

        v = numpy.concatenate((
            v[y1], v[y2], v[y3], v[y4]
        ))

        x = numpy.concatenate((
            x[y1], x[y2], x[y3], x[y4]
        ))

        y = numpy.concatenate((
            y[y1], y[y2], y[y3], y[y4]
        ))

        e = p / ((gamma - 1) * rho)

        m = self.dx * self.dx * rho

        h = self.dx * self.hdx

        fluid = gpa(
            name='fluid', x=x, y=y, u=u, v=v, rho=rho, p=p, e=e, m=m, h=h,
            h0=h
        )

        self.scheme.setup_properties([fluid])
        return [fluid]

    def create_domain(self):
        return DomainManager(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
            periodic_in_x=True, periodic_in_y=True
        )

    def create_scheme(self):
        crk = CRKSPHScheme(
            fluids=['fluid'], dim=2, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=2, has_ghosts=True
        )

        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=0.1, beta=0.1, k=1.2, eps=0.1, g1=0.1, g2=0.2,
            has_ghosts=True)

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.2, alpha1=1.0, alpha2=0.1,
            beta=2.0, update_alpha1=True, update_alpha2=True,
            has_ghosts=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.5,
            g1=0.2, g2=0.4, rsolver=2, interpolation=1, monotonicity=2,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=40, tol=1e-6, has_ghosts=True
        )
        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )
        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )
        s = SchemeChooser(
            default='crksph', crksph=crk, gsph=gsph, adke=adke, mpm=mpm,
            psph=psph, tsph=tsph
        )

        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'crksph':
            s.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=50
            )
        elif self.options.scheme == 'mpm':
            s.configure(kernel_factor=1.2)
            s.configure_solver(dt=dt, tf=tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=dt, tf=tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'gsph':
            s.configure_solver(dt=dt, tf=tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(dt=dt, tf=tf,
                               adaptive_timestep=False, pfreq=50)


if __name__ == "__main__":
    app = KHInstability()
    app.run()
